from __future__ import print_function
"""
Elasticsearch Extension.
"""
import json
import requests
import logging
from tubing import sinks


logger = logging.getLogger('tubing.ext.elasticsearch')


class BulkUpdateBatcherSink(sinks.ProxySink):
    """
    Creates bulk updates for Elastic Search.  This is an object stream handler
    and expects objects with index() and to_dict() methods implemented.
    """
    def __init__(self, sink, batch_size=7500):
        self.sink = sink
        self.batch = []
        self.batch_size = batch_size

    def action(self, doc):
        return json.dumps(dict(
            update=dict(
                _id=doc.index(),
            ),
        ))

    def update(self, doc):
        return json.dumps(dict(
            doc_as_upsert=True,
            doc=doc.to_dict(),
        ))

    def write(self, docs):
        for doc in docs:
            self.batch.append("{}\n{}\n".format(self.action(doc), self.update(doc)))
            if len(self.batch) > self.batch_size:
                batch = ''.join(self.batch[:self.batch_size])
                self.batch = self.batch[self.batch_size:]
                self.sink.write(batch)

    def done(self):
        self.sink.write(''.join(self.batch))
        return super(BulkUpdateBatcherSink, self).done()


class BulkSink(sinks.BaseSink):
    """
    Elastic search bulk writer.
    """
    def __init__(self, base, index_name, doc_type, username, password):
        self.base = base
        self.index_name = index_name
        self.doc_type = doc_type
        self.username = username
        self.password = password

    def _url(self):
        return "{}/{}/{}/_bulk".format(self.base, self.index_name, self.doc_type)

    def _auth(self):
        return self.username, self.password

    def write(self, chunk):
        logger.debug("POSTING: {}".format(chunk))
        resp = requests.post(self._url(), data=chunk, auth=self._auth())
        resp_obj = json.loads(resp.text)
        logger.debug(json.dumps(resp_obj, indent=2))
        # TODO: handle errors



from __future__ import print_function
"""
Tubing sources are defined here. Here's an example of a gziped s3 document with
one JSON object per line::

    bucket = 'mybucket'
    key = 'path/to/myfile.gz'
    source = JSONParser(LineReaderSource(ZlibSource(S3Source(bucket, key))))

    for obj in iter(source.read, ''):
        # do something with obj dict
"""

import json
import logging
import zlib
import gzip


logger = logging.getLogger('tubing.sources')


class ZlibSource(object):
    """
    ZlibSource unzips a gzipped source stream.
    """
    def __init__(self, source):
        self.source = source
        self.dec = zlib.decompressobj(32 + zlib.MAX_WBITS)

    def read(self, amt=None):
        chunk = self.source.read(amt)
        return self.dec.decompress(chunk)


class LineReaderSource(object):
    """
    LineReaderSource allow readline on a source stream.
    """
    def __init__(self, source, chunksize=4096):
        self.source = source
        self.chunksize = chunksize
        self.buffer = []
        self.eof = False

    def read(self, amt=1):
        response = []
        for _ in range(0, amt or 1):
            response.append(self.readline())
            if self.eof:
                break
        return response

    def readline(self):
        """
        Read a single line.
        """
        if self.eof:
            return ''

        output = []
        newbuffer = []
        state = "outputting"
        for chunk in self.buffer:
            if b"\n" in chunk:
                out, buf = chunk.split(b"\n", 1)
                output.append(out)
                newbuffer.append(buf)
                state = "buffering"
            elif state == "outputting":
                output.append(chunk)
            else:
                newbuffer.append(chunk)

        self.buffer = newbuffer
        while state == "outputting":
            chunk = self.source.read(self.chunksize)
            if not chunk:
                # EOF
                self.eof = True
                return b''.join(output)
            if b"\n" in chunk:
                out, buf = chunk.split(b"\n", 1)
                output.append(out)
                self.buffer.append(buf)
                state = "buffering"
            else:
                output.append(chunk)

        return b''.join(output)


class JSONParserSource(object):
    """
    JSONParserSource is not very smart. It expects one raw JSON object per read
    and works well with LineReaderSource for source files with one JSON object
    per line.
    """
    def __init__(self, source, encoding='utf-8'):
        self.source = source
        self.encoding = encoding

    def read(self, amt=1):
        response = []
        for line in self.source.read(amt):
            if line:
                j = line.decode(self.encoding).strip()
                response.append(json.loads(j))
        return response

    def readobj(self):
        return self.read()[0]



from __future__ import print_function
"""
Tubing sinks are targets for streams of data.
"""

import json
import gzip
import logging
try:
    from StringIO import StringIO
except:
    from io import StringIO
from io import BytesIO


logger = logging.getLogger('tubing.sinks')


class BaseSink(object):
    """
    BaseSink is a mixin class for all Sinks providing empty impls of the sink
    interface.
    """

    def write(self, chunk):
        pass

    def done(self):
        pass

    def abort(self):
        pass


class ProxySink(object):
    """
    ProxySink is a mixin class with default implementations of all methods for a
    proxy sink.
    """

    def write(self, chunk):
        return self.sink.write(chunk)

    def done(self):
        return self.sink.done()

    def abort(self):
        return self.sink.abort()


class JSONSerializerSink(ProxySink):
    """
    JSONSerializerSink takes an object stream and serializes it to json.
    delimiter will be inserted after ever record.
    """

    def __init__(self, sink, delimiter="", encoding="utf-8", **kwargs):
        self.sink = sink
        self.delimiter = delimiter
        self.json_kwargs = kwargs
        self.encoding = encoding

    def write(self, objs):
        for obj in objs:
            line = json.dumps(obj, **self.json_kwargs) + self.delimiter
            chunk = line.encode(self.encoding)
            self.sink.write(chunk)


class ZlibSink(ProxySink):
    """
    ZlibSink Gzips the binary input.
    """
    class InnerStream(StringIO):
        def __init__(self, sink):
            self.sink = sink

        def write(self, data):
            self.sink.write(data)

    def __init__(self, sink):
        self.sink = sink
        self.zipfile = gzip.GzipFile("", 'wb', 9, self.InnerStream(sink))

    def write(self, chunk):
        return self.zipfile.write(chunk)

    def done(self):
        self.zipfile.close()
        return super(ZlibSink, self).done()


class BufferedSink(ProxySink):
    """
    BufferedSink buffers output into batches before sending. It expects a text
    stream.
    """
    def __init__(self, sink, batch_size=6000000):
        self.buffer = b""
        self.batch_size = batch_size
        self.sink = sink

    def write(self, chunk):
        """
        write media docs to sink.
        """
        self.buffer += chunk
        while len(self.buffer) >= self.batch_size:
            r = self.sink.write(self.buffer[:self.batch_size])
            self.buffer = self.buffer[self.batch_size:]
            return r

    def done(self):
        """
        finalize sink, flushing the buffer to downstream sink.
        """
        self.sink.write(self.buffer)
        return super(BufferedSink, self).done()


class FileSink(BaseSink):
    """
    FileSink writes data to a file.
    """
    def __init__(self, path):
        self.f = file(path, 'wb')

    def write(self, chunk):
        return self.f.write(chunk)

    def done(self):
        return self.f.close()

    def abort(self):
        return self.f.close()

class BytesIOSink(BytesIO, BaseSink):
    """
    BytesIOSink wraps a BytesIO object for tubing.
    """
    def done(self):
        pass

    def abort(self):
        pass




from __future__ import print_function
"""
S3 Tubing extension.
"""

import boto3
import logging


logger = logging.getLogger('tubing.ext.s3')


def S3Source(bucket, key):
    """
    Create an S3 Source stream.
    """
    s3 = boto3.client('s3')
    response = s3.get_object(Bucket=bucket, Key=key)
    return response['Body']


class S3Sink(object):
    """
    Send MediaDocs to S3. Expects AWS environmental variables to be set.
    """
    def __init__(self, bucket, key):
        """
        Initiate the multipart upload.
        """
        self.s3 = boto3.client('s3')
        self.bucket = bucket
        self.key = key
        logger.debug("Creating upload for {} {}".format(bucket, key))
        upload = self.s3.create_multipart_upload(Bucket=bucket, Key=key)
        self.upload_id = upload["UploadId"]
        # start with 1
        self.part_number = 1
        self.part_info = dict(Parts=[])

    def track_part(self, part_number, part):
        """
        Keep track of all parts so we can finalize multipart upload.
        """
        self.part_info['Parts'].append(dict(
            PartNumber=part_number,
            ETag=part['ETag'],
        ))

    def write(self, chunk):
        """
        Upload a part.
        """
        logger.debug("Posting {}".format(self.part_number))
        part = self.s3.upload_part(
            Bucket=self.bucket,
            Key=self.key,
            PartNumber=self.part_number,
            UploadId=self.upload_id,
            Body=chunk,
        )
        self.track_part(self.part_number, part)
        self.part_number += 1

    def done(self):
        """
        Finalize upload.
        """
        self.s3.complete_multipart_upload(
            Bucket=self.bucket,
            Key=self.key,
            UploadId=self.upload_id,
            MultipartUpload=self.part_info,
        )

    def abort(self):
        """
        Something failed, abort!
        """
        self.s3.abort_multipart_upload(
            Bucket=self.bucket,
            Key=self.key,
            UploadId=self.upload_id,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys

from codecs import open
from datetime import datetime
from setuptools import setup, find_packages

minor_version = "0.0.1"


if sys.argv[-1] == "publish":
    os.system("python setup.py sdist upload")
    sys.exit()


if os.environ.get("TRAVIS_BRANCH") == "release":
    revision = "r" + os.environ.get("TRAVIS_BUILD_NUMBER")
    version = minor_version + revision
else:
    date = datetime.now().strftime(".dev%Y%m%d%H%M%S")
    revision = os.environ.get("REVISION", date)
    version = minor_version + revision

with open("tubing/VERSION", "w", "utf-8") as f:
    f.write(version)


with open("README.rst", "r", "utf-8") as f:
    readme = f.read()


setup(
    name="tubing",
    version=version,
    description="Python I/O pipe utilities",
    long_description=readme,
    author="Bob Corsaro",
    author_email="rcorsaro@gmail.com",
    url="http://github.com/dokipen/tubing",
    package_data={"": ["LICENSE"], "tubing": ["VERSION"]},
    packages=find_packages(exclude='tests'),
    include_package_data=True,
    install_requires=[],
    license="MIT",
    zip_safe=False,
    classifiers=(
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Developers',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ),
    extras_require={
        "build": [
            "nose",
            "unittest2",
        ],
        "s3": [
            "boto3",
        ],
        "elasticsearch": [
            "requests",
        ],
    },
)
